﻿Option Strict On
Option Infer On

Imports DirectShowLib
Imports System.Windows.Forms
Imports System.IO
Imports System.Drawing
Imports System.Runtime.InteropServices
Imports System.Timers
Imports System.Threading

''' <summary>
''' Ermöglicht das Wiedergeben von Video und Audio über DirectShow.
''' </summary>
''' <remarks></remarks>
Public Class MediaPlayer
    Private m_MediaFile As String
    Private m_Handle As IntPtr
    Private m_Loop As Boolean
    Private m_Fullscreen As Boolean
    Private m_Volume As Integer
    Private m_Balance As Integer
    Private m_Speed As Double
    Private loaded As Boolean
    Private oldPos As TimeSpan
    Private WithEvents t As Timers.Timer

    Private fGraph As FilterGraph
    Private filterGraph2 As IFilterGraph2
    Private mediaControl As IMediaControl
    Private mediaEventEx As IMediaEventEx
    Private mediaSeeking As IMediaSeeking
    Private mediaPosition As IMediaPosition
    Private videoWindow As IVideoWindow
    Private basicAudio As IBasicAudio
    Private basicVideo As IBasicVideo

    ''' <summary>
    ''' Tritt ein, wenn die Wiedergabe beendet ist.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Public Event Stopped(sender As Object, e As EventArgs)

    ''' <summary>
    ''' Tritt ein, wenn die Wiedergabe begonnen hat.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Public Event Started(sender As Object, e As EventArgs)

    ''' <summary>
    ''' Tritt ein, wenn sich die Position der Wiedergabe geändert hat.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Public Event PositionChanged(sender As Object, e As EventArgs)

    ''' <summary>
    ''' Die aktuell für die Wiedergabe ausgewählte Mediendatei.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property MediaFile As String
        Get
            Return m_MediaFile
        End Get
        Set(value As String)
            If State <> MediaPlayerState.Stopped Then Me.Stop()
            Unload()
            m_MediaFile = value
        End Set
    End Property

    ''' <summary>
    ''' Ein Handle eines Controls, das als Ziel für die Videowiedergabe verwendet wird.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Handle As IntPtr
        Get
            Return m_Handle
        End Get
        Set(value As IntPtr)
            m_Handle = value
            If State <> MediaPlayerState.Stopped Then SetWindow()
        End Set
    End Property

    ''' <summary>
    ''' Der aktuelle Wiedergabestatus des Mediaplayers.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property State As MediaPlayerState
        Get
            If Not loaded Then Return MediaPlayerState.Stopped
            Dim fState As FilterState
            Dim hResult As Integer = 0
            hResult = mediaControl.GetState(-1, fState)
            DsError.ThrowExceptionForHR(hResult)
            Return CType(fState, MediaPlayerState)
        End Get
    End Property

    ''' <summary>
    ''' Gibt die Abmessungen des gerade ausgewählten Videos zurück. Gibt 0 zurück, wenn eine Audiodatei ausgewählt ist.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property VideoSize As Size
        Get
            Dim s As New Size(0, 0)
            If Not HasVideo Then Return s
            Dim w As Integer = 0
            Dim h As Integer = 0
            basicVideo.get_VideoWidth(w)
            basicVideo.get_VideoHeight(h)
            s.Width = w
            s.Height = h
            Return s
        End Get
    End Property

    ''' <summary>
    ''' Gibt die Wiedergabedauer der Mediendatei bei normaler Geschwindigkeit zurück.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Length As TimeSpan
        Get
            If Not loaded Then Return Nothing
            Dim ticks = 0L
            mediaSeeking.GetDuration(ticks)
            Return New TimeSpan(ticks)
        End Get
    End Property

    ''' <summary>
    ''' Gibt die aktuelle Position der Wiedergabe zurück.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Position As TimeSpan
        Get
            If Not loaded Then Return Nothing
            Dim ticks = 0L
            mediaSeeking.GetCurrentPosition(ticks)
            Return New TimeSpan(ticks)
        End Get
        Set(value As TimeSpan)
            If Not loaded Then Exit Property
            mediaSeeking.SetPositions(New DsLong(value.Ticks), AMSeekingSeekingFlags.AbsolutePositioning, New DsLong(0), AMSeekingSeekingFlags.NoPositioning)
        End Set
    End Property

    ''' <summary>
    ''' Gibt an, ob die Mediendatei ein Video besitzt.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property HasVideo As Boolean
        Get
            If Not loaded Then Return False
            If basicVideo Is Nothing Then Return False
            Dim w As Integer = 0
            Dim h As Integer = 0
            basicVideo.get_VideoWidth(w)
            basicVideo.get_VideoHeight(h)
            Return w > 0 AndAlso h > 0
        End Get
    End Property

    ''' <summary>
    ''' Wenn aktiviert wird die Wiedergabe in einer Dauerschleife ausgeführt, bis die Stop-Methode aufgerufen wird.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property [Loop] As Boolean
        Get
            Return m_Loop
        End Get
        Set(value As Boolean)
            m_Loop = value
        End Set
    End Property

    ''' <summary>
    ''' Gibt an, ob ein Video im Vollbildmodus abgespielt wird.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Fullscreen As Boolean
        Get
            Return m_Fullscreen
        End Get
        Set(value As Boolean)
            m_Fullscreen = value
            SetFullscreen()
        End Set
    End Property

    ''' <summary>
    ''' Gibt die Wiedergabegeschwindigkeit an. 1,0 entspricht dabei normaler Geschwindigkeit.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Speed As Double
        Get
            Return m_Speed
        End Get
        Set(value As Double)
            m_Speed = value
            SetSpeed()
        End Set
    End Property

    ''' <summary>
    ''' Gibt den maximalen Lautstärkepegel an. Konstant.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property MaxVolume As Integer
        Get
            Return 100000
        End Get
    End Property

    ''' <summary>
    ''' Gibt den minimalen Lautstärkepegel an. Konstant.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property MinVolume As Integer
        Get
            Return 1
        End Get
    End Property

    ''' <summary>
    ''' Der aktuelle Lautstärkepegel. Es sind nur Werte zwischen MaxVolume und MinVolume möglich.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Volume As Integer
        Get
            Return Convert.ToInt32(Math.Pow(10, (m_Volume + 10000) / 2000))
        End Get
        Set(value As Integer)
            m_Volume = Convert.ToInt32(2000 * Math.Log10(value)) - 10000
            SetVolume()
        End Set
    End Property

    ''' <summary>
    ''' Stellt den Wert für die alleinige Wiedergabe auf dem linken Audiokanal dar. Konstant.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property BalanceLeft As Integer
        Get
            Return -10000
        End Get
    End Property

    ''' <summary>
    ''' Stellt den Wert für die alleinige Wiedergabe auf dem rechten Audiokanal dar. Konstant.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property BalanceRight As Integer
        Get
            Return 10000
        End Get
    End Property

    ''' <summary>
    ''' Gibt an, wie die Lautstärke auf die Audiokanäle verteilt wird. Es sind nur Werte zwischen BalanceLeft und BalanceRight möglich.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Balance As Integer
        Get
            Return m_Balance
        End Get
        Set(value As Integer)
            m_Balance = Clamp(value, BalanceRight, BalanceLeft)
            SetBalance()
        End Set
    End Property

    Private Function GetOABool(bool As Boolean) As OABool
        If bool Then Return OABool.True
        Return OABool.False
    End Function

    Private Function Clamp(val As Integer, maxVal As Integer, minVal As Integer) As Integer
        If val > maxVal Then Return maxVal
        If val < minVal Then Return minVal
        Return val
    End Function

    Private Sub Unload()
        loaded = False
        Try
            If fGraph IsNot Nothing Then Marshal.ReleaseComObject(fGraph)
        Catch : End Try
    End Sub

    Private Function Load() As Boolean
        If State = MediaPlayerState.Paused Then Return True
        If File.Exists(m_MediaFile) Then
            CreateObjects()
            DsError.ThrowExceptionForHR(filterGraph2.RenderFile(m_MediaFile, Nothing))
            Return True
        End If
        Return False
    End Function

    Private Sub SetWindow()
        If Not HasVideo Then Exit Sub
        Const WMGraphNotify As Integer = &H400 + 13
        Dim hResult As Integer = 0
        hResult = mediaEventEx.SetNotifyWindow(m_Handle, WMGraphNotify, IntPtr.Zero)
        DsError.ThrowExceptionForHR(hResult)
        hResult = videoWindow.put_Owner(m_Handle)
        DsError.ThrowExceptionForHR(hResult)
        hResult = videoWindow.put_WindowStyle(WindowStyle.Child Or WindowStyle.ClipSiblings Or WindowStyle.ClipChildren)
        DsError.ThrowExceptionForHR(hResult)
    End Sub

    Private Sub SetFullscreen()
        If HasVideo Then videoWindow.put_FullScreenMode(GetOABool(m_Fullscreen))
    End Sub

    Private Sub SetVolume()
        If loaded Then basicAudio.put_Volume(m_Volume)
    End Sub

    Private Sub SetBalance()
        If loaded Then basicAudio.put_Balance(m_Balance)
    End Sub

    Private Sub SetSpeed()
        If loaded Then mediaSeeking.SetRate(m_Speed)
    End Sub

    Private Sub CreateObjects()
        fGraph = New FilterGraph
        filterGraph2 = DirectCast(fGraph, IFilterGraph2)
        mediaControl = DirectCast(fGraph, IMediaControl)
        mediaEventEx = DirectCast(fGraph, IMediaEventEx)
        mediaSeeking = DirectCast(fGraph, IMediaSeeking)
        mediaPosition = DirectCast(fGraph, IMediaPosition)
        videoWindow = DirectCast(fGraph, IVideoWindow)
        basicAudio = DirectCast(fGraph, IBasicAudio)
        basicVideo = DirectCast(fGraph, IBasicVideo)
    End Sub

    ''' <summary>
    ''' Startet die Wiedergabe.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Play()
        If State = MediaPlayerState.Playing Then Exit Sub
        loaded = Load()
        If Not loaded Then Exit Sub
        SetWindow()
        SetVolume()
        SetBalance()
        SetSpeed()
        Dim hResult = 0
        hResult = mediaSeeking.SetTimeFormat(TimeFormat.MediaTime)
        DsError.ThrowExceptionForHR(hResult)
        hResult = mediaControl.Run
        DsError.ThrowExceptionForHR(hResult)
        SetFullscreen()
        t.Start()
        RaiseEvent Started(Me, EventArgs.Empty)
    End Sub

    ''' <summary>
    ''' Stoppt die Wiedergabe.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub [Stop]()
        If Not loaded Then Exit Sub
        If State = MediaPlayerState.Stopped Then Exit Sub
        If t.Enabled Then t.Stop()
        Do While t.Enabled
            Thread.Sleep(1)
        Loop
        Dim hResult As Integer = 0
        hResult = mediaControl.Stop
        DsError.ThrowExceptionForHR(hResult)
        Unload()
        RaiseEvent Stopped(Me, EventArgs.Empty)
    End Sub

    ''' <summary>
    ''' Pausiert die Wiedergabe.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub [Pause]()
        If Not loaded Then Exit Sub
        If State <> MediaPlayerState.Playing Then Exit Sub
        Dim hResult As Integer = 0
        hResult = mediaControl.Pause
        DsError.ThrowExceptionForHR(hResult)
    End Sub

    ''' <summary>
    ''' Gibt ein Bitmap-Objekt zurück, das den aktuellen Frame des Videos abbildet.
    ''' Gibt Nothing zurück, wenn die aktuelle Mediendatei kein Video enthält.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetCurrentImage() As Bitmap
        If Not HasVideo Then Return Nothing
        Dim buffer As Integer = 0
        Dim hWnd = IntPtr.Zero
        basicVideo.GetCurrentImage(buffer, hWnd)
        hWnd = Marshal.AllocCoTaskMem(buffer)
        basicVideo.GetCurrentImage(buffer, hWnd)
        Dim bmpInfo As New DirectShowLib.BitmapInfoHeader
        Marshal.PtrToStructure(hWnd, bmpInfo)
        Dim mStep = bmpInfo.ImageSize \ bmpInfo.Height
        Dim bmp As New Bitmap(bmpInfo.Width, bmpInfo.Height, mStep, Imaging.PixelFormat.Format32bppRgb, hWnd)
        bmp.RotateFlip(RotateFlipType.RotateNoneFlipY)
        Marshal.FreeCoTaskMem(hWnd)
        Return bmp
    End Function

    ''' <summary>
    ''' Richtet das Video unter Berücksichtigung des Seitenverhältnisses im angegebenen Rechteck relativ zur Position des Containers zentriert aus.
    ''' </summary>
    ''' <param name="rect"></param>
    ''' <remarks></remarks>
    Public Sub Resize(rect As Rectangle)
        If Not loaded Then Exit Sub
        If Not HasVideo Then Exit Sub
        Dim vSize = Scale(VideoSize.Width, VideoSize.Height, rect.Width, rect.Height, True)
        Dim vLoc As New Point((rect.Width - vSize.Width) \ 2, (rect.Height - vSize.Height) \ 2)
        videoWindow.SetWindowPosition(vLoc.X, vLoc.Y, vSize.Width, vSize.Height)
    End Sub

    Private Function Scale(imgWidth As Integer, imgHeight As Integer, maxWidth As Integer, maxHeight As Integer, stretch As Boolean) As Size
        If imgWidth = 0 OrElse imgHeight = 0 OrElse maxWidth = 0 OrElse maxHeight = 0 Then Return Nothing
        Dim imgRatio = imgWidth / imgHeight
        Dim maxRatio = maxWidth / maxHeight
        If (imgWidth > maxWidth OrElse imgHeight > maxHeight) OrElse stretch Then
            If imgRatio <= maxRatio Then
                imgWidth = Convert.ToInt32(imgWidth / (imgHeight / maxHeight))
                imgHeight = maxHeight
            Else
                imgHeight = Convert.ToInt32(imgHeight / (imgWidth / maxWidth))
                imgWidth = maxWidth
            End If
        End If
        Return New Size(imgWidth, imgHeight)
    End Function

    ''' <summary>
    ''' Erstellt einen neuen Mediaplayer.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        m_Handle = IntPtr.Zero
        m_MediaFile = ""
        loaded = False
        m_Loop = False
        m_Fullscreen = False
        m_Speed = 1
        t = New Timers.Timer(1)
    End Sub

    Private Sub t_Elapsed(sender As Object, e As ElapsedEventArgs) Handles t.Elapsed
        If Not loaded Then
            t.Stop()
            Exit Sub
        End If
        If Position >= Length Then
            If m_Loop Then
                Position = New TimeSpan(0)
            Else
                t.Stop()
                Me.Stop()
                Do While loaded
                    Thread.Sleep(1)
                Loop
            End If
        End If
        Dim pos = Position
        If oldPos <> pos Then
            oldPos = pos
            RaiseEvent PositionChanged(Me, EventArgs.Empty)
        End If
    End Sub
End Class

''' <summary>
''' Stellt einen Wiedergabestatus dar.
''' </summary>
''' <remarks></remarks>
Public Enum MediaPlayerState
    ''' <summary>
    ''' Es wird nichts wiedergegeben.
    ''' </summary>
    ''' <remarks></remarks>
    Stopped
    ''' <summary>
    ''' Die Wiedergabe ist pausiert.
    ''' </summary>
    ''' <remarks></remarks>
    Paused
    ''' <summary>
    ''' Die Wiedergabe läuft gerade.
    ''' </summary>
    ''' <remarks></remarks>
    Playing
End Enum